#!/bin/csh -f
#
# ccbuild.csh: compile and link 'simple' C++ S2PLOT program
#
# ==================================================================
# S2PLOT is freely available for non-commercial and educational use,
# however, it is not public-domain software. The programming library
# and documentation may not be redistributed or sub-licensed in any
# form without permission from Swinburne University of
# Technology. The programming library and sample codes are provided
# without warranty.
#
# We would appreciate it if research outcomes using S2PLOT would
# provide the following acknowledgement:
#
# "Three-dimensional visualisation was conducted with the S2PLOT
# progamming library"
#
# and a reference to 
#
# D.G.Barnes, C.J.Fluke, P.D.Bourke & O.T.Parry, 2006, Publications
# of the Astronomical Society of Australia, 23(2), 82-93.
#
# S2PLOT development was funded through the Researcher Developer
# Grant Scheme at Swinburne University of Technology.
# ==================================================================
#
# original version: David Barnes, June 2007
#
# $Id: ccbuild.csh 5524 2010-11-10 04:53:11Z dbarnes $
#
# usage: ccbuild.csh target
# - compiles 'target.cc' and links to produce executable 'target'

if (!(${?S2PATH})) then
  echo "S2PATH environment variable MUST be set ... please fix and retry."
  exit(-1);
endif
if (! -d $S2PATH || ! -e ${S2PATH}/s2plot.csh) then
  echo "S2PATH is set but invalid: ${S2PATH} ... please fix and retry."
  exit(-1);
endif
source ${S2PATH}/s2plot.csh
if ($status) then
  exit(-1)
endif

set thisdir=$cwd
if ($thisdir == $S2PATH) then
  echo "You must NOT be in directory ${S2PATH} to build your own programs."
  exit(-1);
endif

#set target=$1
set target=`echo $1 | sed -e 's/\.cc$//g'`
set source=${target}.cc
set object=${target}.o
if (! -e ${source}) then
  echo "Cannot find source code: ${source}."
  exit(-1);
endif 

echo "Compiling source code file ${source} ..."
echo $S2CCMPILER ${S2EXTRAINC} -I${S2PATH} $source
$S2CCMPILER ${S2EXTRAINC} -I${S2PATH} $source

echo "Linking object file ${target}.o ..."
echo $S2CCINKER -o $target ${object} -L${S2PATH}/${S2KERNEL} ${S2LINKS} ${SWLINKS} ${GLLINKS} -L${S2X11PATH}/lib${S2LBITS} ${S2FORMSLINK} -lXpm -lX11 ${IMATH} -lm ${XLINKPATH} ${S2EXTRALIB}
$S2CCINKER -o $target ${object} -L${S2PATH}/${S2KERNEL} ${S2LINKS} ${SWLINKS} ${GLLINKS} -L${S2X11PATH}/lib${S2LBITS} ${S2FORMSLINK} -lXpm -lX11 ${IMATH} -lm ${XLINKPATH} ${S2EXTRALIB}

echo "Cleaning up ..."
rm -rf $object

echo "Done!"
