/* s2plot_internal.h: internal "private" data / functions for S2PLOT
 *
 * ==================================================================
 * S2PLOT is freely available for non-commercial and educational use,
 * however, it is not public-domain software. The programming library
 * and documentation may not be redistributed or sub-licensed in any
 * form without permission from Swinburne University of
 * Technology. The programming library and sample codes are provided
 * without warranty.
 *
 * We would appreciate it if research outcomes using S2PLOT would
 * provide the following acknowledgement:
 *
 * "Three-dimensional visualisation was conducted with the S2PLOT
 * progamming library"
 *
 * and a reference to 
 *
 * D.G.Barnes, C.J.Fluke, P.D.Bourke & O.T.Parry, 2006, Publications
 * of the Astronomical Society of Australia, 23(2), 82-93.
 *
 * S2PLOT development was funded through the Researcher Developer
 * Grant Scheme at Swinburne University of Technology.
 * ==================================================================
 *
 * $Id: s2plot_internal.h 5533 2011-05-19 04:23:13Z dbarnes $
 *
 */

#ifndef S2PLOT_INTERNAL_H
#define S2PLOT_INTERNAL_H

#if defined(__cplusplus)
extern "C" {
#endif

#if defined(BUILDING_S2PLOT)

/* world coordinate system */
/* PLEASE NOTE: these are not yet used consistently throughout S2PLOT */
/* THEREFORE DO NOT CHANGE THEM AND EXPECT SENSIBLE RESULTS !!! */
#define _S2XAX 0
#define _S2YAX 1
#define _S2ZAX 2

float _s2devicemin[3]; /* normalised device coord system lower bounds */
float _s2devicemax[3]; /* normalised device coord system upper bounds */

float _s2axismin[3]; /* world coord system axis minima */
float _s2axismax[3]; /* world coord system axis maxima */
float _s2axissgn[3]; /* axis sign - ie. +1 if max > min, -1 otherwise */

int _s2_conelines; /* for private control */

  char _s2_opt_store[3][20]; /* for storing axis labelling options */


#define _S2WORLD2DEVICE(wv, axis) (strlen(_s2_whichscreen) ? (wv) : (_s2devicemin[(axis)] + (_s2devicemax[(axis)]-_s2devicemin[(axis)]) / (_s2axismax[(axis)]-_s2axismin[(axis)]) * ((wv) - _s2axismin[(axis)])))

#define _S2WORLD2DEVICE_SO(wv, axis) (strlen(_s2_whichscreen) ? (wv) : ((_s2devicemax[(axis)]-_s2devicemin[(axis)]) / (_s2axismax[(axis)]-_s2axismin[(axis)]) * ((wv))))

#define _S2DEVICE2WORLD(dv, axis) (strlen(_s2_whichscreen) ? (dv) : (_s2axismin[(axis)] + (_s2axismax[(axis)]-_s2axismin[(axis)]) / (_s2devicemax[(axis)]-_s2devicemin[(axis)]) * ((dv) - _s2devicemin[(axis)])))

#define _S2DEVICE2WORLD_SO(dv, axis) (strlen(_s2_whichscreen) ? (dv) : ((_s2axismax[(axis)]-_s2axismin[(axis)]) / (_s2devicemax[(axis)]-_s2devicemin[(axis)]) * ((dv))))

#define _S2W3RADIUS(val) (strlen(_s2_whichscreen) ? (val) : (sqrt(0.333333 * (powf(_S2WORLD2DEVICE_SO(val, _S2XAX), 2.0) + powf(_S2WORLD2DEVICE_SO(val, _S2YAX), 2.0) + powf(_S2WORLD2DEVICE_SO(val, _S2ZAX), 2.0)))))

/* is a,b,c monotonic ? */
#define _S2MONOTONIC(a,b,c) ((((a)<(b))&&((b)<(c))) || (((a)>(b))&&((b)>(c))) ? 1 : 0)

#define S2_WITHIN(a, b, c) ((fabs((a)-(b)) < (c)) ? 1 : 0)

#define S2_COLOURWITHIN(a, y, c) ( (((a).r-(y).r)*((a).r-(y).r) + ((a).g-(y).g)*((a).g-(y).g) + ((a).b-(y).b)*((a).b-(y).b)) < c*c )

#define S2_XYZWITHIN(a, b, c) ( (((a).x-(b).x)*((a).x-(b).x) + ((a).y-(b).y)*((a).y-(b).y) + ((a).z-(b).z)*((a).z-(b).z)) < c*c )

  //#define S2CLIP(p) (((p).x < _s2axismin[_S2XAX]) || ((p).x > _s2axismax[_S2XAX]) || ((p).y < _s2axismin[_S2YAX]) || ((p).y > _s2axismax[_S2YAX]) || ((p).z < _s2axismin[_S2ZAX]) || ((p).z > _s2axismax[_S2ZAX]))

#define S2INWORLD(p) (_S2MONOTONIC(_s2axismin[_S2XAX], (p).x, _s2axismax[_S2XAX]) && _S2MONOTONIC(_s2axismin[_S2YAX], (p).y, _s2axismax[_S2YAX]) && _S2MONOTONIC(_s2axismin[_S2ZAX], (p).z, _s2axismax[_S2ZAX]))

typedef struct {
  XYZ p;         /* position */
  COLOUR col;    /* colour to draw when not selected */
  COLOUR hilite; /* colour to draw when selected */
  unsigned int id;        /* unique identifier (= glLoadName value) */
  int selected;  /* 0 not selected, 1 selected */
  double dist;   /* distance to camera, used in sort */
  float size;    /* size of handle: approx. the diameter of billboard */
  int texid;     /* texture for unselected draw, -1 for default */
  int hitexid;   /* texture for selected draw, -1 for default */
  char whichscreen[10];
  char VRMLname[32];
} _S2HANDLE;

typedef struct {
  XYZ p;         /* position */
  XYZ str;       /* stretch (direction + magnitude): 0 = no stretch */
  XYZ offset;    /* offset (X and Y only, Z ignored) */
  float aspect;  /* aspect ratio: width to height "on the screen" */
  float size;    /* size of billboard */
  float pa;      /* position angle in radians, clockwise from 12 o'clock */
  COLOUR col;    /* colour */
  unsigned int texid; /* texture identifier */
  float alpha;   /* transparency */
  double dist;   /* distance to camera - used internally */
  char trans;    /* 'o', 's', 't' */
  char whichscreen[10];
  char VRMLname[32];
} _S2BBOARD;

typedef struct {
  GLfloat *vertarray;
  GLfloat *colarray;
  float size;
  unsigned int texid;
  char trans;
  int n;
  char whichscreen[10];
  char VRMLname[32];
} _S2BBSET;

/* transparent 3-vertex facet */
typedef struct {
  XYZ p[3];
  XYZ n[3];
  COLOUR colour[3];
  int trans; /* 'o' = opaque, 't'/'s' = transparent */
  double alpha[3]; /* transparency: 1.0 = default = opaque */
  //int screen; /* 1 = screen coords; 0 = normal (3d) coords */
  char whichscreen[10]; 
  char VRMLname[32];
} _S2FACE3A;

#if defined(S2_3D_TEXTURES)
/* 3d textured polygon, (3d texture, maybe 2d, 4d later) */
typedef struct {
  int nverts;
  XYZ *verts;
  XYZ *texcoords;
  unsigned int texid;
  int trans; /* 'o' = opaque, 't'/'s' = transparent */
  double alpha; /* 1.0 = opaque, 0.0 = totally transparent */
  char whichscreen[10];
  char VRMLname[32];
} _S2TEXPOLY3D;
#endif

/* current character size/height - affects tickmarks and symbols too */
float _s2_charsize;

COLOUR _s2_foreground;

/* current pen color */
#define _S2PENRED (_s2_colormap[_s2_colidx].r)
#define _S2PENGRN (_s2_colormap[_s2_colidx].g)
#define _S2PENBLU (_s2_colormap[_s2_colidx].b)

/* separators for s2iden function */
#define _S2IDENPREFIX ""
#define _S2IDENMIDFIX "; "
#define _S2NULLIDSTR "<unknown>"

/* near & far clipping plane expansion (1.0 = original version) */
float _s2_nearfar_expand;
  /* near and far clipping planes from last iteration */
double _s2_save_near;
double _s2_save_far;

/* current line width */
float _s2_linewidth;

  /* current line style */
  int _s2_linestyle;

/* colormap size, ie. colors are 0 .. (_s2_cmapsize -1) */
int _s2_cmapsize;
/* the colormap itself */
COLOUR *_s2_colormap;
/* the current color index in use */
int _s2_colidx;
/* the current color index range for shading */
int _s2_colr1, _s2_colr2;

#if defined(OLDWEDGE)
/* whether to draw a wedge, and what data range maps to colidx1, colidx2 ? */
int _s2_wedge;
char *_s2_wedgelabel;
float _s2_wedgemin, _s2_wedgemax;
/* fractional location on screen of wedge */
float _s2_wedgex1, _s2_wedgex2, _s2_wedgey1, _s2_wedgey2;
int _s2_wedgecolidx;
#endif

/* arrow properties */
int _s2_arrow_fs = 1; /* 1 = "solid cone"; 2 is not yet supported */
float _s2_arrow_angle = 45.0;
float _s2_arrow_barb = 0.3;

/* title and iden strings */
char *_s2_title;
char *_s2_iden;

/* show annotations? - can be toggled with 'n' or 'N' in interact mode */
int _s2_showannotations;

  /* custom help string for second press of F1 */
  char *_s2_customhelpstr;

#define _S2MAXCOLORS 16384
/* default colormap from PGPlot */
static int _s2_pg_cmapsize = 16;
static COLOUR _s2_pg_colormap[] = {{0.00, 0.00, 0.00}, // black 
				   {1.00, 1.00, 1.00}, // white
				   {1.00, 0.00, 0.00}, // red
				   {0.00, 1.00, 0.00}, // green
				   {0.00, 0.00, 1.00}, // blue
				   {0.00, 1.00, 1.00}, // cyan
				   {1.00, 0.00, 1.00}, // magenta
				   {1.00, 1.00, 0.00}, // yellow
				   {1.00, 0.50, 0.00}, // red+yellow (orange)
				   {0.50, 1.00, 0.00}, // green+yellow
				   {0.00, 1.00, 0.50}, // green+cyan
				   {0.00, 0.50, 1.00}, // blue+cyan
				   {0.50, 0.00, 1.00}, // blue+magenta
				   {1.00, 0.00, 0.50}, // red+magenta
				   {0.33, 0.33, 0.33}, // dark gray
				   {0.66, 0.66, 0.66}};// light gray

/* character spacing in x direction at screen */
#define _S2XSP 18

/* fractional length of ticks for axis labels */
#define _S2TIKFRAC 0.04

#define _S2DEVCAP_CURSOR    0x01
#define _S2DEVCAP_CROSSHAIR 0x02
#define _S2DEVCAP_SELECTION 0x04
#define _S2DEVCAP_NOCOLOR   0x08

/* the devices we support */
typedef struct {
  char *devicename;
  int fullscreen;
  int stereo;
  int advertise;
  char *devicedesc;
} _S2DEVICE;
_S2DEVICE _s2_valid_devices[] = 
  {{"/S2MONO",   0, 0, 1, "Monoscopic, windowed display"},
   {"/S2MONOF",  1, 0, 1, "Monoscopic, full-screen display"},
   {"/S2ACTIV",  0, 1, 1, "Active stereoscopic, windowed display"},
   {"/S2ACTIVF", 1, 1, 1, "Active stereoscopic, full-screen display"},
   {"/S2PASSV",  0, 2, 1, "Passive stereoscopic, windowed display"},
   {"/S2PASSVF", 1, 2, 1, "Passive stereoscopic, full-screen display"},
   {"/S2FISH",   0, 3, 1, "Fisheye projection, windowed display"},
   {"/S2FISHF",  1, 3, 1, "Fisheye projection, full-screen display"},
   {"/S2TRUNCB", 0, 6, 0, "Truncated-base fisheye, windowed display"},
   {"/S2TRUNCBF",1, 6, 0, "Truncated-base fisheye, full-screen display"},
   {"/S2TRUNCT", 0, 7, 0, "Truncated-top fisheye, windowed display"},
   {"/S2TRUNCTF",1, 7, 0, "Truncated-top fisheye, full-screen display"},
   {"/S2WARP",   0, 5, 0, "Warped projection (eg. MirrorDome), windowed"},
   {"/S2WARPF",  1, 5, 0, "Warped projection (eg. MirrorDome), full-screen"},
   {"/S2ANA",    0, 8, 1, "Anaglyph stereo (red(L), blue(R)), windowed"},
   {"/S2ANAF",   1, 8, 1, "Anaglyph stereo (red(L), blue(R)), full-screen"},
   {"/S2DSANA",   0, 9, 1, "Anaglyph stereo (red(L), blue(R)), pre-desaturated, windowed"},
   {"/S2DSANAF",  1, 9, 1, "Anaglyph stereo (red(L), blue(R)), pre-desaturated, full-screen"},
   {"/S2TRIO",   0, 30, 0, "Passive stereo w/ control display"},
   {"/S2INTER",  0, 31, 1, "Interleaved stereoscopic, windowed display"},
   {"/S2INTERF", 1, 31, 1, "Interleaved stereoscopic, full-screen display"},
   {"/S2WPASSV", 0, 32, 0, "Warped passive stereoscopic, windowed display"},
   {"/S2WPASSVF",1, 32, 0, "Warped passive stereoscopic, full-screen display"}};
int _s2_ndevices = 23;
void _s2priv_list_devices();

/* private helper functions */
#if defined(OLDWEDGE)
void _s2priv_imm_drawWedge();
#endif

void _s2priv_drawCrosshair();


/* create a red-X bitmap for texture use */
BITMAP4 *_s2priv_redXtexture(int w, int h);
BITMAP4 *_s2priv_redXtexture3d(int w, int h, int d);

/* setup a texture from a bitmap of given width and height */
  unsigned int _s2priv_setupTexture(int width, int height, BITMAP4 *bitmap,
				    int usemipmaps);
  unsigned int _s2priv_setupTexture3d(int width, int height, int depth,
				      BITMAP4 *bitmap,
				    int usemipmaps);

  void _s2priv_ss2pt(unsigned int itextureID, int usemipmaps);

/* drop a texture from memory */
void _s2priv_dropTexture(unsigned int texid);

/* pointer to callback function */
void (*_s2_callback)(double*, int *);

  /* pointer to callback function that also handles a void data ptr */
  void (*_s2_callbackx)(double*, int*, void *);
  /* pointer to the data */
  void *_s2_callbackx_data;

  /* pointer to user-installed keyboard callback function */
  int (*_s2_user_keys)(unsigned char *);

int _s2_dynamicEnabled;

  /* whether to erase dynamic geometry; controlled via ds2protect()
   * and ds2unprotect calls */
  int _s2_dynamic_erase;

/* switch the geometry lists so new geometry is added to / drawn from 
 * the dynamic lists.  A global flag is set so we know in MakeGeometry
 * to just draw, not create a list for later use.  If erase is non-zero
 * then the dynamic lists will be cleared.  This could be modified in 
 * future to erase some of the lists but not all.
 */
void _s2_startDynamicGeometry(int erase);

/* switch the geometry lists so new geometry is added to / drawn from 
 * the standard (static) lists.
 */
void _s2_endDynamicGeometry();

/* make the dynamic geometry */
//void _s2_MakeDynamicGeometry();


/* status flag: animating or frozen? */
int _s2_animation;

/* count flag: how many times has the spacebar been pressed? */
int _s2_callbackkey;

/* store state for what we are recording (1 = geom + view, 2 = geom +
 * view + image, 3 = continuous geom + view + image), and what frame
 * number are we up to?
 */
int _s2_recstate;
int _s2_recframe;
/* how many places to use in frame strings */
#define _S2FRAMESTRLEN 4
#define _S2GEOMEXT ".s2g"
#define _S2GEOMEXTLEN 4
#define _S2VIEWEXT ".s2v"
#define _S2VIEWEXTLEN 4
#define _S2DUMPEXT ".tga"
#define _S2DUMPEXTLEN 4

int _s2_skip; /* if 1, set back to 0 and leave glut loop */

/* camera control */
int _s2_cameraset; /* should we set the camera position ? */
XYZ _s2_camera_vp; /* view position to set */
XYZ _s2_camera_vu; /* view up       to set */
XYZ _s2_camera_vd; /* view dirn     to set */
void _s2priv_CameraSet(void);
int _s2_transcam_enable; /* is translating the camera allowed (eg. +/- keys) */

/* camera focus control */
int _s2_camexfocus; /* specific focus (mid) point provided ? */
XYZ _s2_camfocus;   /* mid point to use when _s2_camexfocus = 1 */
float _s2_cameraspeed; /* speed / increment size for camera movements */

  /* camera eye sep scale: default 1.0 */
  float _s2_eyesepmul;

  /* is buffer swapping allowed? - turn off for Cocoa use */
  int _s2_bufswap;

  /* should display lists be retained? - turn off for Mac Screensaver use */
  int _s2_retain_lists;

/* callback function for handling keypresses Shift-1 through Shift-0 */

/* pointer to callback function - provides number of key pressed */
void (*_s2_numcb)(int *);

/* pointer to opengl callback function */
void (*_s2_oglcb)();

  /* pointer to remote control callback function */
  int (*_s2_remcb)(char *);

  /* this can be set to 1 in very special circumstances to avoid
   * waiting for a mutex lock in threads */
  int _s2_skiplock; 

  /* what port do remote controllers connect on? */
  int _s2_remoteport;

/* are the handles visible? */
int _s2_handle_vis;

/* draw the handles */
void _s2priv_drawHandles(int doscreen);

/* draw the billboards */
void _s2priv_drawBillboards(void);

  /* draw billboard sets */
  void _s2priv_drawBBsets(void);

  /* add a billboard */
  void _s2priv_bb(XYZ iP, XYZ iStretch, XYZ ioffset, 
		  float aspect, float size, 
		  float pangle, 
		  COLOUR icol, unsigned int iid, float ialpha, 
		  char itrans);

/* handle clicks on handles */
void _s2priv_handleHits(int nhits, GLuint *list);

/* handle clicks on handles IN DOME mode */
void _s2priv_handleDomeHits(int nhits, GLuint *list);

/* pointer to handle callback function */
void (*_s2_handlecallback)(int *);

  /* find the handle with this id. */
  _S2HANDLE *_s2priv_findHandle(GLuint item);

  /* which handle is being dragged?  NULL for none */
  _S2HANDLE *_s2_draghandle;
  int _s2_dragpanel; /* which panel is drag on? */
  int _s2_draghandle_id;
  double _s2_draghandle_basex, _s2_draghandle_basey; // maybe not required?
  double _s2_draghandle_savez; /* z buffer value that is kept */
  GLdouble _s2_dragmodel[16], _s2_dragproj[16];
  GLint _s2_dragview[4];
  int _s2_draghandle_screen; /* if !0 drag is in screen coordinates */

  /* handle drag callback function, gives handle id and new position
   * in world coords */
  void (*_s2_draghandle_callback)(int *, XYZ *);

/* device capabilities */
int _s2_devcap;

/* is the crosshair visible? */
int _s2_crosshair_vis;

/* mouse coords */
int _s2_lastmousex, _s2_lastmousey;

/* are we clipping within the world bounds? */
int _s2_clipping;

/* latex template file name */
char _s2_latex_template_file[200];

/* routines and structure/s for isosurfaces */
int Polygonise(GRIDCELL g,double iso,TRIANGLE *tri);
typedef struct {
  float ***grptr;
  int adim, bdim, cdim;
  int a1, a2, b1, b2, c1, c2;
  float tr[12];
  char trans; /* 'o', 't' or 's' */
  float alpha; /* 1. is opaque */
  float level;
  int resolution;
  float red, green, blue;
  void (*fcol)(float*,float*,float*,float*,float*,float*);
} _S2TRIANGLE_CACHE_DESCR;

typedef struct {
  /* parameters for this cached triangle list */
  float ***grptr;
  _S2TRIANGLE_CACHE_DESCR descr;
  _S2TRIANGLE_CACHE_DESCR cached_descr;
  /* the triangle list */
  int ntri;
  XYZ *trivert; /* ntri * 3 in length */
  XYZ *normals; /* ntri * 3 in length */
  COLOUR *col;  /* ntri in length */
} _S2TRIANGLE_CACHE;
void _s2priv_drawTriangleCache(_S2TRIANGLE_CACHE *cache);
void _s2priv_addToTriangleCache(_S2TRIANGLE_CACHE *cache, XYZ *trivert, 
				//				XYZ cellnormal,
				float *tr, COLOUR col);
void _s2priv_colrfn(float *ix, float *iy, float *iz, 
		    float *r, float *g, float *b);
float _s2priv_colrfn_r, _s2priv_colrfn_g, _s2priv_colrfn_b;
#define _S2ISOSURFACE _S2TRIANGLE_CACHE
_S2ISOSURFACE *_s2_isosurfs;
int _s2_nisosurf;
void _s2priv_generate_isosurface(int isid, int force);
  int _s2_fastsurfaces;

/* volume rendering types, settings */
typedef struct {
  float ***grid;    /* pointer to original data - NOT a copy! */
  int adim, bdim, cdim; /* dimensions of the grid */
  int a1, a2, b1, b2, c1, c2;
  float tr[12];
  char trans; /* 'o', 't' or 's' */
  float datamin, datamax;
  float alphamin, alphamax; /* 1. is opaque */
  float (*alphafn)(float *);
  int ntexts;
  unsigned int *textureids;
  int axis;
  int reverse; /* if non-zero, reverse drawing order */
} _S2VRVOLUME;
_S2VRVOLUME *_s2_volumes;
int _s2_nvol;
  //void _s2priv_load_vr_textures(int vrid, int force);
int _s2priv_ns2cvr(float ***grid, 
		   int adim, int bdim, int cdim,
		   int a1, int a2, int b1, int b2, int c1, int c2,
		   float *tr, char trans, 
		   float datamin, float datamax,
		   float alphamin, float alphamax,
		   float (*ialphafn)(float*));
void _s2priv_load_vr_textures(int vrid, int force, int iaxis);
  int _s2_vralphascaling;

  float _s2_evas_x, _s2_evas_y, _s2_evas_z;

  // internal fn for parametric functions
void _s2priv_s2funuv(float(*fx)(float*,float*), float(*fy)(float*,float*), 
		     float(*fz)(float*,float*), float(*fcol)(float*,float*), 
		     char trans, float(*ialphafn)(float*,float*),
		     float umin, float umax, int uDIV,
		     float vmin, float vmax, int vDIV);

  // toggle fs / constrained window
  void _s2priv_togglefs();

/* wrapper for glColor4f: desaturate in particular modes */
//void _s2priv_glColor4f(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);

  /* s2plot fade in/out routine */
  double _s2_fadetime;
int _s2_fadestatus; /* 0 = start fade-in, 1 = fade-in, 2 = normal running, 
		      * 3 = start fade-out, 4 = fade-out, 5 = exiting */
void _s2_fadeinout(void);

  /* runtime: if > 0 this means quit after so many seconds */
  double _s2_runtime;

  /* event structure */
  #define ASCII_KEY_EVENT 1
  typedef struct {
    double t;       // event time
    int type; // event type
    void *data;    // event data
    int consumed; // has event been used?
  } _S2EVENT;
  _S2EVENT *_s2_events;
  int _s2_nevents;
  void _s2priv_readEvents(char *filename);

  //int _s2_screenco; /* 0 = normal (3d) coords; 1 = screen (2d) coords */
  char _s2_whichscreen[10]; /* can be one or more of l, c, r for screen coordinate drawing; or empty for normal drawing */

int _s2_screenEnabled; /* 0 = drawing normal geometry; 1 = drawing screen geometry */

  char _s2_doingScreen[2]; /* which screen are we currently drawing? */

/* switch the geometry lists so new geometry is added to / drawn from 
 * the screen lists.  A global flag is set so we know in MakeGeometry
 * to just draw, not create a list for later use.  If erase is non-zero
 * then the screen lists will be cleared.
 */
void _s2_startScreenGeometry(int erase);

/* switch the geometry lists so new geometry is added to / drawn from 
 * the previously set list (either static or dynamic).
 */
void _s2_endScreenGeometry();

  /* clear/erase the screen geometry */
void _s2_clearGeometryList();

  /* "command prompt" */
  void _s2priv_initPrompt();
  void _s2priv_finiPrompt();
  //#define S2PROMPT "S2PLOT% "
#define S2PROMPT_LEN 80
#define S2BASEPROMPT_LEN 30
  char _s2prompt[S2PROMPT_LEN];
  int _s2prompt_length;
  char _s2baseprompt[S2BASEPROMPT_LEN];
  float _s2prompt_x, _s2prompt_y;
  
  void _s2_prompt_kbdfunc(unsigned char, int, int);

/* prompt callback function */
  void (*_s2_promptcbx)(char *, void *);
  void *_s2_promptcbx_data;

#include "s2plot_glodef.h"

  /* multi-panel capability */
  typedef struct {
    
    /* layout */
    float x1, y1, x2, y2; /* where is this panel? x,y in [0,1] */

    /* draw this panel? */
    int active;

    /* coordinates */
    float devicemin[3]; /* normalised device coord system lower bounds */
    float devicemax[3]; /* normalised device coord system upper bounds */
    float axismin[3]; /* world coord system axis minima */
    float axismax[3]; /* world coord system axis maxima */
    float axissgn[3]; /* axis sign - ie. +1 if max > min, -1 otherwise */

    /* clipping? */
    int clipping;

    /* camera */
    // make it a one-length array so we can use it as a pointer, but it 
    // is pre-allocated :-)
    //CAMERA camera[1];
    CAMERA *camera;

    /* more coord info */
    XYZ pmin, pmax, pmid;
    double rangemin, rangemax;

    /* various parts of the global option space */
    XYZ *autospin;
    CAMERA camerahome;
    
    /* callbacks */
    void (*callback)(double *, int *);
    void (*callbackx)(double *, int *, void *);
    void *callbackx_data;

    int (*userkeys)(unsigned char *);
    void (*numcb)(int *);
    void (*oglcb)();
    int (*remcb)(char *);

    void (*handlecallback)(int *);
    void (*draghandlecb)(int *, XYZ *);

    void (*promptcb)(char *, void *);
    void *promptcb_data;
    char baseprompt[S2BASEPROMPT_LEN];
    float prompt_x, prompt_y;

    /* dragging state */
    GLdouble dragmodel[16], dragproj[16];
    GLint dragview[4];

    int GL_listindex; /* for GL list of static geom */
    
    /* "current" geometry */
    int nball    ; BALL    *ball;
    int nballt   ; BALLT   *ballt;
    int ndisk    ; DISK    *disk;
    int ncone    ; CONE    *cone;
    int ndot     ; DOT     *dot;
    int nline    ; LINE    *line;
    int nface3   ; FACE3   *face3;
    int nface4   ; FACE4   *face4;
    int nface4t  ; FACE4T  *face4t;
    int nlabel   ; LABEL   *label;
    int nhandle  ; _S2HANDLE  *handle;
    int nbboard  ; _S2BBOARD  *bboard;
    int nbbset   ; _S2BBSET *bbset;
    int nface3a  ; _S2FACE3A  *face3a;
    int ntrdot   ; TRDOT   *trdot;
#if defined(S2_3D_TEXTURES)
    int ntexpoly3d ; _S2TEXPOLY3D *texpoly3d;
#endif
    
    /* "static" geometry */
    int nball_s    ; BALL    *ball_s;
    int nballt_s   ; BALLT   *ballt_s;
    int ndisk_s    ; DISK    *disk_s;
    int ncone_s    ; CONE    *cone_s;
    int ndot_s     ; DOT     *dot_s;
    int nline_s    ; LINE    *line_s;
    int nface3_s   ; FACE3   *face3_s;
    int nface4_s   ; FACE4   *face4_s;
    int nface4t_s  ; FACE4T  *face4t_s;
    int nlabel_s   ; LABEL   *label_s;
    int nhandle_s  ; _S2HANDLE  *handle_s;
    int nbboard_s  ; _S2BBOARD  *bboard_s;
    int nbbset_s   ; _S2BBSET *bbset_s;
    int nface3a_s  ; _S2FACE3A  *face3a_s;
    int ntrdot_s   ; TRDOT   *trdot_s;
#if defined(S2_3D_TEXTURES)
    int ntexpoly3d_s ; _S2TEXPOLY3D *texpoly3d_s;
#endif
    
    /* "dynamic" geometry */
    int nball_d    ; BALL    *ball_d;
    int nballt_d   ; BALLT   *ballt_d;
    int ndisk_d    ; DISK    *disk_d;
    int ncone_d    ; CONE    *cone_d;
    int ndot_d     ; DOT     *dot_d;
    int nline_d    ; LINE    *line_d;
    int nface3_d   ; FACE3   *face3_d;
    int nface4_d   ; FACE4   *face4_d;
    int nface4t_d  ; FACE4T  *face4t_d;
    int nlabel_d   ; LABEL   *label_d;
    int nhandle_d  ; _S2HANDLE  *handle_d;
    int nbboard_d  ; _S2BBOARD  *bboard_d;
    int nbbset_d   ; _S2BBSET *bbset_d;
    int nface3a_d  ; _S2FACE3A  *face3a_d;
    int ntrdot_d   ; TRDOT   *trdot_d;
#if defined(S2_3D_TEXTURES)
    int ntexpoly3d_d ; _S2TEXPOLY3D *texpoly3d_d;
#endif
    
    /* "screen" geometry */
    int nball_scr    ; BALL    *ball_scr;
    int nballt_scr   ; BALLT   *ballt_scr;
    int ndisk_scr    ; DISK    *disk_scr;
    int ncone_scr    ; CONE    *cone_scr;
    int ndot_scr     ; DOT     *dot_scr;
    int nline_scr    ; LINE    *line_scr;
    int nface3_scr   ; FACE3   *face3_scr;
    int nface4_scr   ; FACE4   *face4_scr;
    int nface4t_scr  ; FACE4T  *face4t_scr;
    int nlabel_scr   ; LABEL   *label_scr;
    int nhandle_scr  ; _S2HANDLE  *handle_scr;
    int nbboard_scr  ; _S2BBOARD  *bboard_scr;
    int nbbset_scr   ; _S2BBSET *bbset_scr;
    int nface3a_scr  ; _S2FACE3A  *face3a_scr;
    int ntrdot_scr   ; TRDOT   *trdot_scr;
#if defined(S2_3D_TEXTURES)
    int ntexpoly3d_scr ; _S2TEXPOLY3D *texpoly3d_scr;
#endif
    
  } S2PLOT_PANEL;
  S2PLOT_PANEL *_s2_panels;
  int _s2_npanels;
  int _s2_activepanel;

  void _s2priv_pushGlobalsToPanel(int panelid);
  void _s2priv_pushPanelToGlobals(int panelid);
  void _s2priv_drawActiveFrame(int active);

  COLOUR _s2_activepanelframecolour;
  COLOUR _s2_panelframecolour;
  float _s2_panelframewidth;

  // what are the screen constraints?
  float _s2_scr_x1, _s2_scr_x2, _s2_scr_y1, _s2_scr_y2;

  // background image texture id: if < 0, not available
  int _s2_bg_texid;
  void _s2_drawBGimage(void);
  void _s2_blankPanel(void);
  int _s2_bg_clear;

  int _s2_nVRMLnames;
  char **_s2_VRMLnames;
  int _s2_currVRMLidx;

  void writeVRML20(void);
  void doVRMLdots(FILE *fp, int start, int end);
  void doVRMLlines(FILE *fp, int start, int end);
  void doVRMLface4(FILE *fp, int start, int end);
  void doVRMLface4t(FILE *fp, int start, int end);
  void doVRMLface3(FILE *fp, int start, int end);
  void doVRMLface3a(FILE *fp, int start, int end);
  void doVRMLball(FILE *fp, int start, int end);
  void doVRMLcone(FILE *fp, int start, int end);
  void doVRMLcylinder(FILE *fp, int start, int end);
  void doVRMLlabel(FILE *fp, int start, int end);
  void doVRMLbboard(FILE *fp, int start, int end);
  void doVRMLhandle(FILE *fp, int start, int end);

  void writeASYLATEX(void);
  void doASYLATEXdots(FILE *fp, int start, int end);
  void doASYLATEXlines(FILE *fp, int start, int end);
  void doASYLATEXface4(FILE *fp, int start, int end);
  void doASYLATEXface4t(FILE *fp, int start, int end);
  void doASYLATEXface3(FILE *fp, int start, int end);
  void doASYLATEXlabel(FILE *fp, int start, int end);

  void writePRC(void);

#endif

  /* are we using an ati card? (requires no multisampling) */
  int _s2x_ati;

  /* what is the top scan line number? */
  int _s2x_ywinpos;

/* cache of textures (generally used in callbacks) */
int _s2_ctext_count;
typedef struct {
  int width, height; // "user" width and height
  int width2, height2; // "internal" width and height (powers of two)
  float wfrac, hfrac; // width and height fraction for texture coordinates
                      // when width2 > width &/ height2 > height
  BITMAP4 *bitmap;
  GLuint id;
  int depth, depth2; // "user" and "internal" depth for 3d textures
} _S2CACHEDTEXTURE;
_S2CACHEDTEXTURE *_s2_ctext;

#if defined(__cplusplus)
} // extern "C" {
#endif

#endif /* S2PLOT_INTERNAL_H */
