/* sock.h: socket interface for S2PLOT library
 *
 * ==================================================================
 * S2PLOT is freely available for non-commercial and educational use,
 * however, it is not public-domain software. The programming library
 * and documentation may not be redistributed or sub-licensed in any
 * form without permission from Swinburne University of
 * Technology. The programming library and sample codes are provided
 * without warranty.
 *
 * We would appreciate it if research outcomes using S2PLOT would
 * provide the following acknowledgement:
 *
 * "Three-dimensional visualisation was conducted with the S2PLOT
 * progamming library"
 *
 * and a reference to
 *
 * D.G.Barnes, C.J.Fluke, P.D.Bourke & O.T.Parry, 2006, Publications
 * of the Astronomical Society of Australia, 23(2), 82-93.
 *
 * S2PLOT development was funded through the Researcher Developer
 * Grant Scheme at Swinburne University of Technology.
 * ==================================================================
 *
 * original version: Willem van Straten, c. 1996
 * contributed to S2PLOT: 20080904
 *
 * $Id: sock.h 5274 2008-09-04 01:52:14Z dbarnes $
 *
 */

#ifndef __DADA_SOCK_H
#define __DADA_SOCK_H

#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif
  
  int sock_getname (char* self, int length, int alias);
  
  int sock_create (int* port);
  int sock_accept (int fd);
  
  int sock_open (const char* host, int port);
  int sock_close (int fd);
  
  int sock_read (int fd, void* buf, size_t size);
  int sock_write (int fd, const void* buf, size_t size);
  
  int sock_ready (int fd, int* to_read, int* to_write, float timeout);
  int sock_tm_read (int fd, void* buf, size_t size, float timeout);
  int sock_tm_write (int fd, void* buf, size_t size, float timeout);
  
  int sock_block (int fd);
  int sock_nonblock (int fd);
  
#ifdef __cplusplus
}
#endif

#endif /* SOCK_H */

